/*   EXTRAITS DE LA LICENCE
	Copyright CEA, contributeurs : Luc BILLARD et Damien
	CALISTE, laboratoire L_Sim, (2001-2005)
  
	Adresse ml :
	BILLARD, non joignable par ml ;
	CALISTE, damien P caliste AT cea P fr.

	Ce logiciel est un programme informatique servant  visualiser des
	structures atomiques dans un rendu pseudo-3D. 

	Ce logiciel est rgi par la licence CeCILL soumise au droit franais et
	respectant les principes de diffusion des logiciels libres. Vous pouvez
	utiliser, modifier et/ou redistribuer ce programme sous les conditions
	de la licence CeCILL telle que diffuse par le CEA, le CNRS et l'INRIA 
	sur le site "http://www.cecill.info".

	Le fait que vous puissiez accder  cet en-tte signifie que vous avez 
	pris connaissance de la licence CeCILL, et que vous en avez accept les
	termes (cf. le fichier Documentation/licence.fr.txt fourni avec ce logiciel).
*/

/*   LICENCE SUM UP
	Copyright CEA, contributors : Luc BILLARD et Damien
	CALISTE, laboratoire L_Sim, (2001-2005)

	E-mail address:
	BILLARD, not reachable any more ;
	CALISTE, damien P caliste AT cea P fr.

	This software is a computer program whose purpose is to visualize atomic
	configurations in 3D.

	This software is governed by the CeCILL  license under French law and
	abiding by the rules of distribution of free software.  You can  use, 
	modify and/ or redistribute the software under the terms of the CeCILL
	license as circulated by CEA, CNRS and INRIA at the following URL
	"http://www.cecill.info". 

	The fact that you are presently reading this means that you have had
	knowledge of the CeCILL license and that you accept its terms. You can
	find a copy of this licence shipped with this software at Documentation/licence.en.txt.
*/
#include "panelBrowser.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <gtk/gtk.h>

#include <visu_gtk.h>
#include <visu_object.h>
#include <visu_tools.h>
#include <visu_rendering.h>
#include <visu_basic.h>
#include <support.h>
#include <gtk_main.h>
#include <extraGtkFunctions/gtk_dumpDialogWidget.h>
#include <gtk_openGLWidget.h>
#include <renderingBackend/visu_windowInterface.h>
#include <gtk_renderingWindowWidget.h>

static GtkWidget *panelBrowser;
static GtkTreeStore *treeStoreFiles;
static GtkTreeModel *treeStoreFilesFilter;
enum
  {
    COLUMN_BOOLEAN, /* Rendered or not. */
    COLUMN_NAME,  /* File name in locale of the file system. */
    COLUMN_NAME_UTF8,  /* File name in UTF8 (to view it it GTK). */
    COLUMN_DATA,  /* File info (size). */
    COLUMN_ACTIVE, /* A boolean to set if this file is filtered or not. */
    COLUMN_FILE_KIND, /* An integer that correspond to the file type. */
    N_COLUMNS
  };
static GtkWidget *fileTree;
static GtkWidget *hboxProgress;
static GtkWidget *progressbar;
static GtkWidget *scrolledwindow1;
static GtkWidget *labelDirectory;
static gboolean flagListLoaded;
static gboolean flagWorkInProgress;
static gboolean flagBarIsShown;
/* A NULL terminated array of directories names, currently printed. */
static gchar **currentBrowseredDirectory;
/* The string common to all paths in currentBrowseredDirectory. */
static gchar *commonBrowseredDirectory;

GtkWidget *entryFilterBrowser;
enum
  {
    PANEL_BROWSER_COLUMN_FILTER_LABEL,
    PANEL_BROWSER_COLUMN_FILTER_ID,
    PANEL_BROWSER_N_COLUMN_FILTER
  };
GtkListStore *panelBrowserListFilter;
GtkWidget *panelBrowserComboFilter;
gint currentComboFilterValue;

static GtkWidget *imagePlay;
static GtkWidget *imageStop;
static GtkWidget *buttonPlayStop;
static GtkWidget *spinDelay;

GtkWidget *buttonPrevious;
GtkWidget *buttonNext;
GtkWidget *buttonSelectAll;
GtkWidget *buttonUnselectAll;
GtkWidget *buttonDirectory;
static GtkWidget *buttonRecurse;
GtkWidget *radioGoAround, *radioGoOnce, *radioGoAndBack;
static int currentBrowseDirection;
static GtkTreePath *startBrowsePath;

GtkWidget *buttonDumpAll;
/* GtkProgressBar *progressBarDump; */


static GtkWidget *createInteriorBrowser();
static void panelBrowserBrowse_directory();
static void addParsedDirectory(int commonPathLen, const gchar *root,
			       GDir *gdir, gboolean recurse,
			       int nbKind, int *nbPattern, GPatternSpec ***pattern,
			       int *nbFiles, gchar nbFilesLabel[36]);
static gboolean showProgressBar(gpointer data);
/* Load and render the given file. The iter is given in the filter
   model. */
static gboolean browserLoad(gchar *filename, gchar* fileUTF8, int kind,
			    GtkTreeIter *iter, int nSet);
/* For the given iter load it. The iter must be given
   in the filter model. */
static gboolean navigateInFiles(GtkTreePath *path, GtkTreeIter *iterSelected);
static void selectFile(GtkTreePath *path, GtkTreeIter *iter);

gboolean gtk_tree_model_get_iter_last(GtkTreeModel *model, GtkTreeIter *last, GtkTreePath **path);
void panelBrowserSet_labelCurrentDir();

/* Callbacks */
static void onCheckFile(GtkCellRendererToggle *cell_renderer, gchar *path,
			gpointer user_data);
static void treeViewActivated(GtkTreeView *treeview, GtkTreePath *path,
			      GtkTreeViewColumn *col, gpointer user_data);
static void navigateClicked(GtkButton *button, gpointer data);
void checkFiles(GtkButton *button, gpointer data);
static void onDirectoryClicked(GtkButton *button, gpointer data);
static void onPlayStopClicked(GtkToggleButton *button _U_, gpointer data _U_);
static void stopPlayStop(gpointer data);
static void onSpinDelayChangeValue(GtkSpinButton *spinbutton, gpointer user_data);
void onDumpButtonClicked(GtkButton *button, gpointer user_data);
void filterBrowserChanged(GtkEditable *entry, gpointer user_data);
void abortDumpAll(GtkButton *button, gpointer data);
void updateDumpAllProgressBar(gpointer data);
static void onRenderingChanged(GObject *obj, RenderingMethod *method,
			       gpointer data);
gboolean panelBrowserIsIterVisible(GtkTreeModel *model, GtkTreeIter *iter, gpointer data);
void onComboFilterChanged(GtkComboBox *combo, gpointer data);
static void onRefreshDir(GtkButton *button, gpointer data);
static void onBrowserEnter(ToolPanel *toolPanel, gpointer data);
gboolean checkFile(GtkTreeModel *model, GtkTreePath *path,
		   GtkTreeIter *iter, gpointer data);
static void onParseAbortClicked(GtkButton *button, gpointer data);
static void onRecurseToggled(GtkToggleButton *toggle, gpointer data);



ToolPanel* panelBrowser_init()
{
  char *cl = _("Browser");
  char *tl = _("Browser");
  GtkCellRenderer *renderer;
  GtkTreeViewColumn *column;

  /* Create the list that store the filters. */
  panelBrowserListFilter = gtk_list_store_new(PANEL_BROWSER_N_COLUMN_FILTER,
					      G_TYPE_STRING,
					      G_TYPE_INT);


  panelBrowser = toolPanelNew_withIconFromPath("Panel_browser", cl, tl,
					       "stock-browser_20.png");
  if (!panelBrowser)
    return (ToolPanel*)0;
  gtk_container_add(GTK_CONTAINER(panelBrowser), createInteriorBrowser());
  toolPanelSet_dockable(TOOL_PANEL(panelBrowser), TRUE);
						 
  /* Create the tree structure. */
  treeStoreFiles = gtk_tree_store_new(N_COLUMNS,
				      G_TYPE_BOOLEAN,
				      G_TYPE_STRING,
				      G_TYPE_STRING,
				      G_TYPE_STRING,
				      G_TYPE_BOOLEAN,
				      G_TYPE_INT);
  /* Use the COLUMN_ACTIVE as a flag to hide or not the rows that 
     don't match the filter value. */
  treeStoreFilesFilter = gtk_tree_model_filter_new(GTK_TREE_MODEL(treeStoreFiles), NULL);
  gtk_tree_model_filter_set_visible_func(GTK_TREE_MODEL_FILTER(treeStoreFilesFilter),
					 panelBrowserIsIterVisible, (gpointer)0,
					 (GtkDestroyNotify)0);
/*   gtk_tree_model_filter_set_visible_column(GTK_TREE_MODEL_FILTER(treeStoreFilesFilter), COLUMN_ACTIVE); */

  /* Add a new row to the model */
  gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(treeStoreFiles),
				       COLUMN_NAME_UTF8, GTK_SORT_ASCENDING);

  renderer = gtk_cell_renderer_toggle_new ();
  g_signal_connect(G_OBJECT(renderer), "toggled",
		   G_CALLBACK(onCheckFile), (gpointer)0);
  column = gtk_tree_view_column_new_with_attributes ("Render",
						     renderer,
						     "active", COLUMN_BOOLEAN,
						     NULL);
  gtk_tree_view_append_column (GTK_TREE_VIEW (fileTree), column);
  renderer = gtk_cell_renderer_text_new ();
  column = gtk_tree_view_column_new_with_attributes ("Filename",
						     renderer,
						     "markup", COLUMN_NAME_UTF8,
						     NULL);
  gtk_tree_view_append_column (GTK_TREE_VIEW (fileTree), column);
/*   renderer = gtk_cell_renderer_text_new (); */
/*   column = gtk_tree_view_column_new_with_attributes ("Data", */
/* 						     renderer, */
/* 						     "text", COLUMN_DATA, */
/* 						     NULL); */
/*   gtk_tree_view_append_column (GTK_TREE_VIEW (fileTree), column); */

  gtk_tree_selection_set_mode(gtk_tree_view_get_selection(GTK_TREE_VIEW(fileTree)),
			      GTK_SELECTION_SINGLE);

  /* Initialise the values. */
  currentBrowseredDirectory = (gchar**)0;
  commonBrowseredDirectory = (gchar*)0;
  flagListLoaded = FALSE;
  currentBrowseDirection = BROWSER_NEXT;

  /* Create the callbacks of all the sensitive widgets. */
  g_signal_connect(G_OBJECT(panelBrowser), "page-entered",
		   G_CALLBACK(onBrowserEnter), (gpointer)0);

  return TOOL_PANEL(panelBrowser);
}

static GtkWidget *createInteriorBrowser()
{
  GtkWidget *vbox1;

  /* include from glade */
  GtkWidget *image2;
  GtkWidget *image3;
  GtkWidget *image1;
  GtkWidget *hbox2;
  GtkWidget *label4;
  GtkObject *spinDelay_adj;
  GtkWidget *label1;
  GtkWidget *alignment1;
  GtkWidget *hbox5;
  GtkWidget *image4;
  GtkWidget *refreshDirButton;
  GtkWidget *vBoxWaiting;
  GtkTooltips *tooltips;

  GtkWidget *label, *wd;
  GtkWidget *hbox, *vbox;
  GtkWidget *hrule;

  GtkCellRenderer *renderer;

  GSList *radiobuttonCycle_group;

  vbox1 = gtk_vbox_new(FALSE, 0);

  tooltips = gtk_tooltips_new ();

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox1), hbox, FALSE, FALSE, 0);
  /* This is a label in the browser panel to introduce the
     entry that allows to enter a filter for files shown. */
  label = gtk_label_new(_("Filter: "));
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 2);
  entryFilterBrowser = gtk_entry_new();
  gtk_entry_set_editable(GTK_ENTRY(entryFilterBrowser), TRUE);
  gtk_entry_set_text(GTK_ENTRY(entryFilterBrowser), "*");
  gtk_box_pack_start(GTK_BOX(hbox), entryFilterBrowser, TRUE, TRUE, 0);
  panelBrowserComboFilter = gtk_combo_box_new_with_model(GTK_TREE_MODEL(panelBrowserListFilter));
  renderer = gtk_cell_renderer_text_new();
  gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(panelBrowserComboFilter), renderer, FALSE);
  gtk_cell_layout_add_attribute(GTK_CELL_LAYOUT(panelBrowserComboFilter), renderer,
				"text", PANEL_BROWSER_COLUMN_FILTER_LABEL);
  gtk_box_pack_start(GTK_BOX(hbox), panelBrowserComboFilter, FALSE, FALSE, 2);

  /*******************/
  /* File lists part */
  /*******************/
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox1), hbox, TRUE, TRUE, 0);
  
  vbox = gtk_vbox_new (FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 2);

  buttonDirectory = gtk_button_new ();
  gtk_box_pack_start (GTK_BOX (vbox), buttonDirectory, FALSE, FALSE, 2);
  gtk_tooltips_set_tip (tooltips, buttonDirectory,
			_("Choose a different directory."), NULL);

  image3 = gtk_image_new_from_stock ("gtk-open", GTK_ICON_SIZE_BUTTON);
  gtk_container_add (GTK_CONTAINER (buttonDirectory), image3);

  refreshDirButton = gtk_button_new ();
  gtk_tooltips_set_tip (tooltips, refreshDirButton,
			_("Rescan current directory."), NULL);
  gtk_box_pack_start(GTK_BOX(vbox), refreshDirButton, FALSE, FALSE, 2);
  image3 = create_pixmap ((GtkWidget*)0, "stock-refresh-dir_20.png");
  gtk_container_add(GTK_CONTAINER(refreshDirButton), image3);
  g_signal_connect(G_OBJECT(refreshDirButton), "clicked",
		   G_CALLBACK(onRefreshDir), (gpointer)0);

  buttonRecurse = gtk_toggle_button_new();
  gtk_tooltips_set_tip (tooltips, buttonRecurse,
			_("Read directories recursively."), NULL);
  gtk_box_pack_start(GTK_BOX(vbox), buttonRecurse, FALSE, FALSE, 2);
  image3 = gtk_image_new_from_stock (GTK_STOCK_ZOOM_IN, GTK_ICON_SIZE_BUTTON);;
  gtk_container_add(GTK_CONTAINER(buttonRecurse), image3);
  g_signal_connect(G_OBJECT(buttonRecurse), "toggled",
		   G_CALLBACK(onRecurseToggled), (gpointer)0);

#if GTK_MINOR_VERSION > 5
  label = gtk_label_new(_("Directory tools:"));
  gtk_label_set_angle(GTK_LABEL(label), 90.);
  gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 5);
#endif

  vBoxWaiting = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), vBoxWaiting, TRUE, TRUE, 0);

  scrolledwindow1 = gtk_scrolled_window_new (NULL, NULL);
  gtk_box_pack_start(GTK_BOX(vBoxWaiting), scrolledwindow1, TRUE, TRUE, 0);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledwindow1),
				 GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
  gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrolledwindow1),
				      GTK_SHADOW_ETCHED_IN);

  fileTree = gtk_tree_view_new ();
  gtk_tooltips_set_tip (tooltips, fileTree,
			_("Double click on a file to render it."), NULL);
  gtk_container_add (GTK_CONTAINER (scrolledwindow1), fileTree);
  gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (fileTree), FALSE);

  hboxProgress = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vBoxWaiting), hboxProgress, FALSE, FALSE, 0);

  progressbar = gtk_progress_bar_new();
  gtk_box_pack_start(GTK_BOX(hboxProgress), progressbar, TRUE, TRUE, 0);

  wd = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
  gtk_box_pack_start(GTK_BOX(hboxProgress), wd, FALSE, FALSE, 2);
  g_signal_connect(G_OBJECT(wd), "clicked",
		   G_CALLBACK(onParseAbortClicked), (gpointer)0);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 2);

  buttonUnselectAll = gtk_button_new();
  gtk_tooltips_set_tip(tooltips, buttonUnselectAll, _("Unselect all files."), NULL);
  gtk_box_pack_start(GTK_BOX(vbox), buttonUnselectAll, FALSE, FALSE, 1);
  image2 = create_pixmap ((GtkWidget*)0, "stock-unselect-all_20.png");
  gtk_container_add(GTK_CONTAINER(buttonUnselectAll), image2);

  buttonSelectAll = gtk_button_new();
  gtk_tooltips_set_tip(tooltips, buttonSelectAll, _("Select all files."), NULL);
  gtk_box_pack_start(GTK_BOX (vbox), buttonSelectAll, FALSE, FALSE, 1);
  image2 = create_pixmap ((GtkWidget*)0, "stock-select-all_20.png");
  gtk_container_add (GTK_CONTAINER (buttonSelectAll), image2);

  alignment1 = gtk_alignment_new(0.5, 0.5, 0., 0.);
  gtk_box_pack_start(GTK_BOX(vbox), alignment1, TRUE, TRUE, 0);
  buttonDumpAll = gtk_button_new();
  gtk_container_add(GTK_CONTAINER(alignment1), buttonDumpAll);
  image4 = gtk_image_new_from_stock("gtk-save-as", GTK_ICON_SIZE_BUTTON);
  gtk_container_add(GTK_CONTAINER(buttonDumpAll), image4);
  gtk_tooltips_set_tip(tooltips, buttonDumpAll,
		       _("Export the rendering of all selected files"
			 " in other formats."), NULL);

  buttonNext = gtk_button_new ();
  gtk_box_pack_end(GTK_BOX (vbox), buttonNext, FALSE, FALSE, 1);
  gtk_tooltips_set_tip(tooltips, buttonNext,
		       _("Render the next selected file."), NULL);
  image1 = gtk_image_new_from_stock ("gtk-go-down", GTK_ICON_SIZE_BUTTON);
  gtk_container_add (GTK_CONTAINER (buttonNext), image1);

  buttonPrevious = gtk_button_new ();
  gtk_box_pack_end(GTK_BOX (vbox), buttonPrevious, FALSE, FALSE, 1);
  gtk_tooltips_set_tip(tooltips, buttonPrevious,
		       _("Render the previous selected file."), NULL);
  image2 = gtk_image_new_from_stock ("gtk-go-up", GTK_ICON_SIZE_BUTTON);
  gtk_container_add (GTK_CONTAINER (buttonPrevious), image2);

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox1), hbox, FALSE, FALSE, 0);
  
  label = gtk_label_new(_("<span size=\"smaller\">Current dir.: </span>"));
  gtk_label_set_use_markup(GTK_LABEL(label), TRUE);
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
  
  labelDirectory = gtk_label_new("");
  gtk_label_set_use_markup(GTK_LABEL(labelDirectory), TRUE);
  gtk_misc_set_alignment(GTK_MISC(labelDirectory), 0., 0.5);
#if GTK_MINOR_VERSION > 5
  gtk_label_set_ellipsize(GTK_LABEL(labelDirectory), PANGO_ELLIPSIZE_START);
#endif
  gtk_box_pack_start(GTK_BOX(hbox), labelDirectory, TRUE, TRUE, 0);
  panelBrowserSet_labelCurrentDir();

  /***************/
  /* Action part */
  /***************/
  alignment1 = gtk_alignment_new(0.5, 0.5, 0.3, 0);
  gtk_box_pack_start(GTK_BOX(vbox1), alignment1, FALSE, FALSE, 2);

  hrule = gtk_hseparator_new();
  gtk_container_add(GTK_CONTAINER(alignment1), hrule);

  hbox2 = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (vbox1), hbox2, FALSE, TRUE, 1);

  alignment1 = gtk_alignment_new(0.5, 0.5, 1.0, 0.0);
  gtk_alignment_set_padding(GTK_ALIGNMENT(alignment1), 0, 0, 0, 5);
  gtk_box_pack_start(GTK_BOX(hbox2), alignment1, FALSE, FALSE, 0);

  hbox5 = gtk_hbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(alignment1), hbox5);
  
  radioGoAround = gtk_radio_button_new(NULL);
  gtk_radio_button_set_group(GTK_RADIO_BUTTON(radioGoAround), (GSList*)0);
  radiobuttonCycle_group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(radioGoAround));
  gtk_box_pack_start(GTK_BOX(hbox5), radioGoAround, FALSE, FALSE, 0);
  image1 = create_pixmap((GtkWidget*)0, "stock-go-around.png");
  gtk_container_add(GTK_CONTAINER(radioGoAround), image1);

  radioGoOnce = gtk_radio_button_new(NULL);
  gtk_radio_button_set_group(GTK_RADIO_BUTTON(radioGoOnce), radiobuttonCycle_group);
  radiobuttonCycle_group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(radioGoOnce));
  gtk_box_pack_start(GTK_BOX(hbox5), radioGoOnce, FALSE, FALSE, 0);
  image1 = create_pixmap((GtkWidget*)0, "stock-go-once.png");
  gtk_container_add(GTK_CONTAINER(radioGoOnce), image1);

  radioGoAndBack = gtk_radio_button_new(NULL);
  gtk_radio_button_set_group(GTK_RADIO_BUTTON(radioGoAndBack), radiobuttonCycle_group);
  radiobuttonCycle_group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(radioGoAndBack));
  gtk_box_pack_start(GTK_BOX(hbox5), radioGoAndBack, FALSE, FALSE, 0);
  image1 = create_pixmap((GtkWidget*)0, "stock-go-and-back.png");
  gtk_container_add(GTK_CONTAINER(radioGoAndBack), image1);

  label4 = gtk_label_new (_("Play at "));
  gtk_box_pack_start (GTK_BOX (hbox2), label4, TRUE, TRUE, 0);
  gtk_misc_set_alignment (GTK_MISC (label4), 1, 0.5);

  spinDelay_adj = gtk_adjustment_new (500, 10, 10000, 25, 100, 100);
  spinDelay = gtk_spin_button_new (GTK_ADJUSTMENT (spinDelay_adj), 1, 0);
  gtk_box_pack_start (GTK_BOX (hbox2), spinDelay, FALSE, TRUE, 0);
  gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (spinDelay), TRUE);

  /* Units: milliseconds */
  label1 = gtk_label_new (_(" ms"));
  gtk_box_pack_start (GTK_BOX (hbox2), label1, TRUE, TRUE, 0);
  gtk_misc_set_alignment (GTK_MISC (label1), 0, 0.5);

  buttonPlayStop = gtk_toggle_button_new();
  gtk_tooltips_set_tip (tooltips, buttonPlayStop, _("Cycle through the selected files at the given rate."), NULL);
  gtk_box_pack_start (GTK_BOX (hbox2), buttonPlayStop, FALSE, FALSE, 5);

  hbox5 = gtk_hbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (buttonPlayStop), hbox5);

#if GTK_MINOR_VERSION < 5
  imagePlay = create_pixmap ((GtkWidget*)0, "stock_media-play.png");
  imageStop = create_pixmap ((GtkWidget*)0, "stock_media-stop.png");
#else
  imagePlay = gtk_image_new_from_stock(GTK_STOCK_MEDIA_PLAY, GTK_ICON_SIZE_BUTTON);
  imageStop = gtk_image_new_from_stock(GTK_STOCK_MEDIA_STOP, GTK_ICON_SIZE_BUTTON);
#endif
  gtk_box_pack_start (GTK_BOX (hbox5), imagePlay, TRUE, TRUE, 0);
  gtk_box_pack_start (GTK_BOX (hbox5), imageStop, TRUE, TRUE, 0);

  g_signal_connect(G_OBJECT(fileTree), "row-activated",
		   G_CALLBACK(treeViewActivated), (gpointer)0);
  g_signal_connect(G_OBJECT(buttonPrevious), "clicked",
		   G_CALLBACK(navigateClicked), (gpointer)BROWSER_PREVIOUS);
  g_signal_connect(G_OBJECT(buttonNext), "clicked",
		   G_CALLBACK(navigateClicked), (gpointer)BROWSER_NEXT);
  g_signal_connect(G_OBJECT(buttonSelectAll), "clicked",
		   G_CALLBACK(checkFiles), (gpointer)1);
  g_signal_connect(G_OBJECT(buttonUnselectAll), "clicked",
		   G_CALLBACK(checkFiles), (gpointer)0);
  g_signal_connect(G_OBJECT(buttonDirectory), "clicked",
		   G_CALLBACK(onDirectoryClicked), (gpointer)0);
  g_signal_connect(G_OBJECT(buttonPlayStop), "toggled",
		   G_CALLBACK(onPlayStopClicked), (gpointer)0);
  g_signal_connect(G_OBJECT(spinDelay), "value-changed",
		   G_CALLBACK(onSpinDelayChangeValue), (gpointer)buttonPlayStop);
  g_signal_connect(G_OBJECT(buttonDumpAll), "clicked",
		   G_CALLBACK(onDumpButtonClicked), (gpointer)0);
  g_signal_connect(VISU_INSTANCE, "renderingChanged",
		   G_CALLBACK(onRenderingChanged), (gpointer)0);
  g_signal_connect(G_OBJECT(entryFilterBrowser), "changed",
		   G_CALLBACK(filterBrowserChanged), (gpointer)0);
  g_signal_connect(G_OBJECT(panelBrowserComboFilter), "changed",
		   G_CALLBACK(onComboFilterChanged), (gpointer)0);

  gtk_widget_show_all(vbox1);
  gtk_widget_hide(hboxProgress);

  gtk_widget_hide(imageStop);
  return vbox1;
}

/*************/
/* Callbacks */
/*************/

void filterBrowserChanged(GtkEditable *entry, gpointer user_data _U_)
{
  GPatternSpec *pattern;
  gboolean valid;
  GtkTreeIter iter;
  gboolean match;
  gchar *fileUTF8;

/*   fprintf(stderr, "'%s'\n", gtk_entry_get_text(GTK_ENTRY(entry))); */

  pattern = g_pattern_spec_new(gtk_entry_get_text(GTK_ENTRY(entry)));

  valid = gtk_tree_model_get_iter_first(GTK_TREE_MODEL(treeStoreFiles), &iter);
  while (valid)
    {
      gtk_tree_model_get(GTK_TREE_MODEL(treeStoreFiles), &iter,
			 COLUMN_NAME_UTF8, &fileUTF8, -1);
      match = g_pattern_match_string(pattern, fileUTF8);
      gtk_tree_store_set(treeStoreFiles, &iter,
			 COLUMN_ACTIVE, match, -1);
      valid = gtk_tree_model_iter_next(GTK_TREE_MODEL(treeStoreFiles), &iter);
    }
  gtk_tree_model_filter_refilter(GTK_TREE_MODEL_FILTER(treeStoreFilesFilter));
}

static void onCheckFile(GtkCellRendererToggle *cell_renderer _U_, gchar *path,
			gpointer user_data _U_)
{
  GtkTreeIter iter, iterList;
  gboolean checked;
  GtkTreePath *currentPath;

  currentPath = gtk_tree_path_new_from_string(path);
  gtk_tree_model_get_iter(GTK_TREE_MODEL(treeStoreFilesFilter), &iter, currentPath);
  gtk_tree_model_get(GTK_TREE_MODEL(treeStoreFilesFilter), &iter,
		     COLUMN_BOOLEAN, &checked, -1);
  gtk_tree_model_filter_convert_iter_to_child_iter(GTK_TREE_MODEL_FILTER(treeStoreFilesFilter),
						   &iterList, &iter);
  gtk_tree_store_set(treeStoreFiles, &iterList,
		     COLUMN_BOOLEAN, !checked, -1);
  gtk_tree_path_free(currentPath);
}

static void treeViewActivated(GtkTreeView *treeview _U_, GtkTreePath *path,
			      GtkTreeViewColumn *col _U_, gpointer user_data _U_)
{
  GtkTreeIter iter, childIter, parentIter;
  gboolean checked;
  gchar* filename, *utf8;
  int kind, nSet;
  GtkTreePath *currentPath;

  gtk_tree_model_get_iter(GTK_TREE_MODEL(treeStoreFilesFilter), &iter, path);
  gtk_tree_model_filter_convert_iter_to_child_iter
    (GTK_TREE_MODEL_FILTER(treeStoreFilesFilter), &childIter, &iter);
  gtk_tree_model_get(GTK_TREE_MODEL(treeStoreFiles), &childIter,
		     COLUMN_NAME, &filename,
		     COLUMN_NAME_UTF8, &utf8,
		     COLUMN_BOOLEAN, &checked,
		     COLUMN_FILE_KIND, &kind, -1);
  if (!checked)
    gtk_tree_store_set(treeStoreFiles, &childIter,
		       COLUMN_BOOLEAN, TRUE, -1);
  if (gtk_tree_model_iter_parent(GTK_TREE_MODEL(treeStoreFiles),
				 &parentIter, &childIter))
    {
      /* We can't use the given path to get the child number since
	 this number may have been filtered. */
      currentPath = gtk_tree_model_get_path(GTK_TREE_MODEL(treeStoreFiles),
					    &childIter);
      nSet = *(gtk_tree_path_get_indices(currentPath) + 1) + 1;
      gtk_tree_path_free(currentPath);
    }
  else
    nSet = 0;

  DBG_fprintf(stderr, "Panel Browser: double click asks for loading (%s).\n",
	      filename);
  browserLoad(filename, utf8, kind, &iter, nSet);
  g_free(filename);
  g_free(utf8);
}
static void navigateClicked(GtkButton *button _U_, gpointer data)
{
  GtkTreeIter iter;
  GtkTreePath *path;
  gboolean res;

  res = panelBrowserGet_nextSelected(&path, &iter, GPOINTER_TO_INT(data));
  if (!res)
    return;

  selectFile(path, &iter);
  res = navigateInFiles(path, &iter);
  gtk_tree_path_free(path);
}
void checkFiles(GtkButton *button _U_, gpointer data)
{
  gtk_tree_model_foreach(GTK_TREE_MODEL(treeStoreFilesFilter), checkFile, data);
}
static void onRenderingChanged(GObject *obj _U_, RenderingMethod *method,
			       gpointer data _U_)
{
  int nb, i;
  GtkTreeIter iter;

  DBG_fprintf(stderr, "Panel Browser : catch the 'renderingChanged' signal.\n");
  /* Clear everything. */
  gtk_list_store_clear(panelBrowserListFilter);
  gtk_tree_store_clear(treeStoreFiles);
  flagListLoaded = FALSE;
  if (!method)
    return;

  /* Update the filters. */
  nb = renderingMethodGet_nbFileType(method);
  if (nb == 1)
    gtk_widget_set_sensitive(panelBrowserComboFilter, FALSE);
  else
    gtk_widget_set_sensitive(panelBrowserComboFilter, TRUE);
  for (i = 0; i < nb; i++)
    {
      gtk_list_store_append(panelBrowserListFilter, &iter);
      gtk_list_store_set(panelBrowserListFilter, &iter,
			 PANEL_BROWSER_COLUMN_FILTER_LABEL,
			 renderingMethodGet_fileTypeName(method, i),
			 PANEL_BROWSER_COLUMN_FILTER_ID, i,
			 -1);
    }
  if (nb > 1)
    {
      gtk_list_store_append(panelBrowserListFilter, &iter);
      gtk_list_store_set(panelBrowserListFilter, &iter,
			 PANEL_BROWSER_COLUMN_FILTER_LABEL, "Similar name files",
			 PANEL_BROWSER_COLUMN_FILTER_ID, nb,
			 -1);
    }
  gtk_combo_box_set_active(GTK_COMBO_BOX(panelBrowserComboFilter), 0);
  currentComboFilterValue = 0;
}
static void onRefreshDir(GtkButton *button _U_, gpointer data _U_)
{
  panelBrowserBrowse_directory((gchar*)0);
}
gboolean playSelectedFiles(gpointer data _U_)
{
  GtkTreeIter iter;
  gboolean res, load;
  GtkTreePath *path;

  g_return_val_if_fail(startBrowsePath, FALSE);

  /* We try to get the next iter. */
  res = panelBrowserGet_nextSelected(&path, &iter, currentBrowseDirection);
  if (!res)
    return FALSE;

  /* We follow the cycle policy to select the file. */
  load = TRUE;
  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(radioGoAndBack)))
    {
      if (!gtk_tree_path_compare(path, startBrowsePath))
	{
	  DBG_fprintf(stderr, "Panel Browser : One round is done, applyng policy.\n");
	  if (currentBrowseDirection == BROWSER_PREVIOUS)
	    currentBrowseDirection = BROWSER_NEXT;
	  else
	    {
	      currentBrowseDirection = BROWSER_PREVIOUS;
	      gtk_tree_path_free(path);  
	      res = panelBrowserGet_nextSelected(&path, &iter,
						 currentBrowseDirection);
	      g_return_val_if_fail(res, FALSE);
	    }
	}
    }
  else if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(radioGoOnce)))
    {
      if (!gtk_tree_path_compare(path, startBrowsePath))
	{
	  DBG_fprintf(stderr, "Panel Browser : One round is done, applyng policy.\n");
	  res = FALSE;
	  load = FALSE;
	}
    }

  if (load)
    {
      /* We advance the select one. */
      selectFile(path, &iter);
      /* We render it. */
      res = navigateInFiles(path, &iter);
      g_return_val_if_fail(res, FALSE);
    }

  gtk_tree_path_free(path);  

  return res;
}

static void onPlayStopClicked(GtkToggleButton *button, gpointer data _U_)
{
  GtkTreeIter startIter;
  gboolean res, checked;
  gulong *playCallbackId;

  DBG_fprintf(stderr, "Panel Browser: toggle on play button.\n");
/*   if (!gtk_tree_model_get_iter_first(GTK_TREE_MODEL(panelBrowserListFilter), &startIter)) */
/*     return; */

  if (gtk_toggle_button_get_active(button))
    {
      /* Button has been pushed. */
      checked = FALSE;
      res = panelBrowserGet_currentSelected(&startBrowsePath, &startIter);
      if (res)
	gtk_tree_model_get(GTK_TREE_MODEL(treeStoreFilesFilter), &startIter,
			   COLUMN_BOOLEAN, &checked, -1);
      if (!res || !checked)
	{
	  res = panelBrowserGet_nextSelected(&startBrowsePath, &startIter,
					     BROWSER_NEXT);
	  if (!res)
	    {
	      startBrowsePath = gtk_tree_path_new();
	      /* No file selected. */
	      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(buttonPlayStop), FALSE);
	      return;
	    }
	  selectFile(startBrowsePath, &startIter);
	}
      navigateInFiles(startBrowsePath, &startIter);

      /* Launch play */
      gtk_widget_hide(imagePlay);
      gtk_widget_show(imageStop);

      currentBrowseDirection = BROWSER_NEXT;
      playCallbackId = g_malloc(sizeof(gulong));
      *playCallbackId =
	g_timeout_add_full(G_PRIORITY_HIGH_IDLE,
			   (gint)gtk_spin_button_get_value(GTK_SPIN_BUTTON(spinDelay)),
			   playSelectedFiles, (gpointer)0,
			   stopPlayStop);
      g_object_set_data(G_OBJECT(button), "playCallbackId",
			(gpointer)playCallbackId);
      DBG_fprintf(stderr, " | start play\n");
    }
  else
    {
      /* Stop play */
      playCallbackId = (gulong*)g_object_get_data(G_OBJECT(button), "playCallbackId");
      if (playCallbackId)
	g_source_remove(*playCallbackId);
      DBG_fprintf(stderr, " | stop play\n");
    }
}
void stopPlayStop(gpointer data _U_)
{
  gtk_widget_hide(imageStop);
  gtk_widget_show(imagePlay);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(buttonPlayStop), FALSE);
  g_return_if_fail(startBrowsePath);
  gtk_tree_path_free(startBrowsePath);
}
static void onSpinDelayChangeValue(GtkSpinButton *spinbutton _U_, gpointer user_data)
{
  GtkTreePath *startPath;
  GtkToggleButton *button;
  gulong *playCallbackId;

  g_return_if_fail(GTK_IS_TOGGLE_BUTTON(user_data));
  button = GTK_TOGGLE_BUTTON(user_data);

  if (gtk_toggle_button_get_active(button))
    {
      playCallbackId = (gulong*)g_object_get_data(G_OBJECT(button), "playCallbackId");
      g_return_if_fail(playCallbackId);

      startPath = gtk_tree_path_copy(startBrowsePath);
      /* Stop playing at this rate */
      g_source_remove(*playCallbackId);
      startBrowsePath = startPath;
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(buttonPlayStop), TRUE);
    }
}
void onDumpButtonClicked(GtkButton *button _U_, gpointer user_data _U_)
{
  GtkWidget *dump;
  GtkProgressBar *progressBarDump;
  char *filename;
  DumpType *format;
  int res;
  GError *error;
  GString *buffer;
  GtkButton *abort;
  char *chr, *chr2;
  int goodPattern, flagAbort, i;
  gint response;
  GString *fileNumbered;
  GtkTreeIter iter;
  gboolean valid, checked, errors;
  GtkTreePath *currentPath;

  dump = dumpDialog_new(toolPanelGet_visuData(TOOL_PANEL(panelBrowser)),
			toolPanelGet_containerWindow(TOOL_PANEL(panelBrowser)),
			_("foo%02d.png"));
  
  do
    {
      response = gtk_dialog_run(GTK_DIALOG(dump));

      if (response != GTK_RESPONSE_ACCEPT)
	{
	  gtk_widget_destroy(dump);
	  return;
	}

      filename = dumpDialogGet_fileName(DUMP_DIALOG(dump));
      format = dumpDialogGet_dumpType(DUMP_DIALOG(dump));
      g_return_if_fail(filename && format);

      DBG_fprintf(stderr, "Panel Browser : dump all returns this filename"
		  " pattern '%s' (format : %s)\n", filename, format->fileType->description);

      res = 0;
      buffer = g_string_new(_("Dumping all selected files to images,"));
      g_string_append_printf(buffer, _(" format '%s'.\n\n"), format->fileType->description);
      /* Verify the name is regular */
      if (g_pattern_match_simple("*%0?d*", filename))
	{
	  chr = strchr(filename, '%');
	  if ((int)*(chr + 2) <= '0' || (int)*(chr + 2) > '9')
	    {
	      goodPattern = 0;
	      g_string_append_printf(buffer, _("Error! The numbering pattern is"
					       " wrong.\n"));
	    }
	  else
	    {
	      chr2 = strchr(chr + 1, '%');
	      if (chr2)
		{
		  goodPattern = 0;
		  g_string_append_printf(buffer,
					 _("Error! Only one '%s' character"
					   " is allowed in the file name.\n"), "%%");
		}
	      else
		goodPattern = 1;
	    }
	}
      else
	{
	  goodPattern = 0;
	  g_string_append_printf(buffer, _("Error! Missing pattern in the filename.\n"));
	}
      if (!goodPattern)
	{
	  g_string_append_printf(buffer,
				 _("\nHelp : you must specify '%s' in"
				   " the filename, where 'x' is a number [|1;9|]."
				   " This allows V_Sim to number the dumped"
				   " files.\n\n For example, with a pattern like this"
				   " : 'foo%s.pdf', dumped files will be named"
				   " : foo00.pdf, foo01.pdf..."),
				 "%%0xd", "%%02d");
	  visuGtkRaise_warning(_("Exporting files"), buffer->str, (GtkWindow*)0);
	  g_string_free(buffer, TRUE);
	}
    }
  while (!goodPattern);

  error = (GError*)0;
  abort = dumpDialogGet_cancelButton(DUMP_DIALOG(dump));
  progressBarDump = dumpDialogGet_progressBar(DUMP_DIALOG(dump));
  dumpDialogStart(DUMP_DIALOG(dump));

  g_signal_connect (G_OBJECT(abort), "clicked",
		    G_CALLBACK(visuDumpAbort), (gpointer)&flagAbort);
  g_signal_connect (G_OBJECT(abort), "clicked",
		    G_CALLBACK(abortDumpAll), (gpointer)progressBarDump);

  gtk_progress_bar_set_fraction(progressBarDump, 0.);
  fileNumbered = g_string_new("");
  i = 0;
  flagAbort = 0;

  /* We write it only once to avoid blinking effects. */
  gtk_progress_bar_set_text(progressBarDump,
			    _("Waiting for generating image in memory..."));
  visuGtkWait();

  valid = gtk_tree_model_get_iter_first(GTK_TREE_MODEL(treeStoreFilesFilter), &iter);
  while (valid && !error && !flagAbort)
    {
      checked = FALSE;
      gtk_tree_model_get(GTK_TREE_MODEL(treeStoreFilesFilter), &iter,
			 COLUMN_BOOLEAN, &checked, -1);
      if (checked)
	{
	  g_string_append_printf(buffer, _("Write to file %d ..."), i);

	  currentPath = gtk_tree_model_get_path(GTK_TREE_MODEL(treeStoreFilesFilter), &iter);
	  res = navigateInFiles(currentPath, &iter);
	  gtk_tree_path_free(currentPath);
      
	  if (!res)
	    error = g_error_new(VISU_ERROR_DUMP, DUMP_ERROR_FILE,
				  _("Can't find next file to export."));
	  else
	    {
	      g_string_printf(fileNumbered, filename, i);
	      DBG_fprintf(stderr, "Panel browser : write '%s'\n",
			  fileNumbered->str);

	      res = (int)visuRenderingWindowDump
		(RENDERING_WINDOW(currentRenderingWindow),
		 format, fileNumbered->str,
		 dumpDialogGet_widthValue(DUMP_DIALOG(dump)),
		 dumpDialogGet_heightValue(DUMP_DIALOG(dump)),
		 &error,
		 updateDumpAllProgressBar, (gpointer)progressBarDump); 
	      if (!res)
		{
		  g_string_append_printf(buffer, _(" error\n"));
		  errors = TRUE;
		}
	      else
		g_string_append_printf(buffer, _(" OK\n"));
	    }
	  i++;
	}
      valid = gtk_tree_model_iter_next(GTK_TREE_MODEL(treeStoreFilesFilter), &iter);
    }
  if (error)
    {
      visuGtkRaise_warning(_("Exporting files"), error->message, (GtkWindow*)0);
      g_error_free(error);
    }
  g_string_free(fileNumbered, TRUE);

  gtk_widget_destroy(dump);
}
void updateDumpAllProgressBar(gpointer data)
{
  gdouble val;
  gdouble percentage;
  gdouble nEle;

  g_return_if_fail(GTK_PROGRESS_BAR(data));

  nEle = (gdouble)gtk_tree_model_iter_n_children(GTK_TREE_MODEL(treeStoreFilesFilter), NULL);
  val = gtk_progress_bar_get_fraction(GTK_PROGRESS_BAR(data));
/*   if (((int)(val * nEle)) % 100 == 1) */
    gtk_progress_bar_set_text(GTK_PROGRESS_BAR(data), "");

  percentage = val + 0.01 / nEle;
  if (percentage > 1.0)
    percentage = 1.0;
  if (percentage < 0.)
    percentage = 0.;
  gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(data), percentage);
  visuGtkWait();
}
void abortDumpAll(GtkButton *button _U_, gpointer data)
{
  gtk_progress_bar_set_text(GTK_PROGRESS_BAR(data),
			    _("Abortion request, please wait..."));
}





/******************/
/* Public methods */
/******************/

gboolean panelBrowserGet_currentSelected(GtkTreePath **path, GtkTreeIter *iterSelected)
{
  GtkTreeSelection *selection;
  gboolean res;
  GtkTreeModel *model;

  g_return_val_if_fail(path && iterSelected, FALSE);

  if (!gtk_tree_model_get_iter_first(GTK_TREE_MODEL(treeStoreFilesFilter), iterSelected))
    return FALSE;

  selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(fileTree));
  res = gtk_tree_selection_get_selected(selection, &model, iterSelected);

  if (res)
    *path = gtk_tree_model_get_path(model, iterSelected);
  return res;
}

gboolean panelBrowserGet_nextSelected(GtkTreePath **path, GtkTreeIter *iterSelected,
				      int direction)
{
  GtkTreeSelection *selection;
  GtkTreePath *currentPath, *firstPath;
  GtkTreeIter iter, child, parent;
  gboolean res, checked;
  GtkTreeModel *model;
  int loopComplete;

  g_return_val_if_fail(path && iterSelected &&
		       (direction == BROWSER_NEXT ||
			direction == BROWSER_PREVIOUS), FALSE);

  *path = (GtkTreePath*)0;

  if (!gtk_tree_model_get_iter_first(GTK_TREE_MODEL(treeStoreFilesFilter), &iter))
    return FALSE;

  selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(fileTree));
  res = gtk_tree_selection_get_selected(selection, &model, &iter);

  /* Cas o rien n'est slectionn */
  if (!res)
    {
      if (direction == BROWSER_NEXT)
	res = gtk_tree_model_get_iter_last(GTK_TREE_MODEL(treeStoreFilesFilter),
					   &iter, (GtkTreePath**)0);
      else if (GPOINTER_TO_INT(direction) == BROWSER_PREVIOUS)
	res = gtk_tree_model_get_iter_first(GTK_TREE_MODEL(treeStoreFilesFilter),
					    &iter);
    }
  g_return_val_if_fail(res, FALSE);

  currentPath = gtk_tree_model_get_path(GTK_TREE_MODEL(treeStoreFilesFilter), &iter);
  firstPath = gtk_tree_path_copy(currentPath);
  do
    {
      if (direction == BROWSER_NEXT)
	{
	  /* If the current iter has child, we select the first. */
	  if (gtk_tree_model_iter_has_child(GTK_TREE_MODEL(treeStoreFilesFilter),
					    &iter))
	    {
	      gtk_tree_model_iter_children(GTK_TREE_MODEL(treeStoreFilesFilter),
					   &child, &iter);
	      iter = child;
	      gtk_tree_path_free(currentPath);
	      currentPath =
		gtk_tree_model_get_path(GTK_TREE_MODEL(treeStoreFilesFilter), &iter);
	    }
	  /* If the current iter has no child, we go next. */
	  else
	    {
	      /* Avance d'un pas dans la liste */
	      gtk_tree_path_next(currentPath);
	      /* Recupre l'itration avance d'un pas */
	      child = iter;
	      res = gtk_tree_model_get_iter(GTK_TREE_MODEL(treeStoreFilesFilter),
					    &iter, currentPath);
	      /* In case, there's no next iter. */
	      if (!res)
		{
		  /* If the iter has no parent we go top. */
		  if (!gtk_tree_model_iter_parent
		      (GTK_TREE_MODEL(treeStoreFilesFilter), &parent, &child))
		    {
		      gtk_tree_model_get_iter_first
			(GTK_TREE_MODEL(treeStoreFilesFilter), &iter);
		      gtk_tree_path_free(currentPath);
		      currentPath =
			gtk_tree_model_get_path
			(GTK_TREE_MODEL(treeStoreFilesFilter), &iter);
		    }
		  /* If the iter has parent, we go next parent. */
		  else
		    {
		      iter = parent;
		      gtk_tree_path_free(currentPath);
		      currentPath = gtk_tree_model_get_path
			(GTK_TREE_MODEL(treeStoreFilesFilter), &iter);
		      gtk_tree_path_next(currentPath);
		      res = gtk_tree_model_get_iter
			(GTK_TREE_MODEL(treeStoreFilesFilter), &iter, currentPath);
		      if (!res)
			{
			  gtk_tree_model_get_iter_first
			    (GTK_TREE_MODEL(treeStoreFilesFilter), &iter);
			  gtk_tree_path_free(currentPath);
			  currentPath = gtk_tree_model_get_path
			    (GTK_TREE_MODEL(treeStoreFilesFilter), &iter);
			}
		    }
		}
	    }
	}
      else if (direction == BROWSER_PREVIOUS)
	{
	  /* Recule d'un pas dans la liste */
	  res = gtk_tree_path_prev(currentPath);
	  /* Recupre l'itration recule d'un pas */
	  child = iter;
	  res = res && gtk_tree_model_get_iter(GTK_TREE_MODEL(treeStoreFilesFilter),
					       &iter, currentPath);
	  /* If this prev iter does not exist. */
	  if (!res)
	    {
	      /* In the case the current iter has a parent, we select it. */
	      if (gtk_tree_model_iter_parent(GTK_TREE_MODEL(treeStoreFilesFilter),
					     &parent, &child))
		{
		  iter = parent;
		  gtk_tree_path_free(currentPath);
		  currentPath = gtk_tree_model_get_path
		    (GTK_TREE_MODEL(treeStoreFilesFilter), &iter);
		}
	      /* Si ce path n'existe pas ou 
		 si cette iteration n'existe pas on va  la fin. */
	      else
		{
		  gtk_tree_path_free(currentPath);
		  res = gtk_tree_model_get_iter_last
		    (GTK_TREE_MODEL(treeStoreFilesFilter), &iter, &currentPath);
		  if (!res)
		    {
		      g_warning("Panel browser : impossible to find"
				" the end of the list.\n");
		      gtk_tree_path_free(currentPath);
		      gtk_tree_path_free(firstPath);
		      return FALSE;
		    }
		}
	    }
	  /* If the previous iter has children, we select last. */
	  else if (gtk_tree_model_iter_has_child(GTK_TREE_MODEL(treeStoreFilesFilter),
						 &iter))
	    {
	      parent = iter;
	      res = gtk_tree_model_iter_nth_child
		(GTK_TREE_MODEL(treeStoreFilesFilter), &iter, &parent,
		 gtk_tree_model_iter_n_children(GTK_TREE_MODEL(treeStoreFilesFilter),
						&parent) - 1);
	      if (!res)
		{
		  g_warning("Panel browser : impossible to find"
			    " the last child of this iter.\n");
		  gtk_tree_path_free(currentPath);
		  gtk_tree_path_free(firstPath);
		  return FALSE;
		}
	      gtk_tree_path_free(currentPath);
	      currentPath = gtk_tree_model_get_path
		(GTK_TREE_MODEL(treeStoreFilesFilter), &iter);
	    }
	}

      /* Rcupre les donnes pour l'itration avance d'un pas. */
      checked = FALSE;
      gtk_tree_model_get(GTK_TREE_MODEL(treeStoreFilesFilter), &iter,
			 COLUMN_BOOLEAN, &checked, -1);
/*       fprintf(stderr, "%d '%s'\n", nbSteps, filename); */
      loopComplete = !gtk_tree_path_compare(firstPath, currentPath);
    }
  while ( !checked && !loopComplete);
  gtk_tree_path_free(firstPath);
  if (!loopComplete)
    {
      *path = currentPath;
      *iterSelected = iter;
      return TRUE;
    }
  else
    {
      gtk_tree_path_free(currentPath);
      return FALSE;
    }
}

static void selectFile(GtkTreePath *path, GtkTreeIter *iter)
{
  GtkTreeSelection *selection;

  g_return_if_fail(path && iter);

  /* Rend l'iter slectionn */
  selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(fileTree));
  gtk_tree_selection_select_iter(selection, iter);
  gtk_tree_view_scroll_to_cell(GTK_TREE_VIEW(fileTree), path, NULL, FALSE, 0., 0.);
}
static gboolean navigateInFiles(GtkTreePath *path, GtkTreeIter *iterSelected)
{
  gboolean res;
  gchar *filename, *utf8;
  int kind, nSet;
  GtkTreeIter parentIter, iter;
  GtkTreePath *currentPath;

  g_return_val_if_fail(path && iterSelected, FALSE);

  /* We select the given iter. */
  selectFile(path, iterSelected);

  /* Load the new selected file */
  gtk_tree_model_filter_convert_iter_to_child_iter
    (GTK_TREE_MODEL_FILTER(treeStoreFilesFilter), &iter, iterSelected);
  if (gtk_tree_model_iter_parent(GTK_TREE_MODEL(treeStoreFiles),
				 &parentIter, &iter))
    {
      /* We can't use the given path to get the child number since
	 this number may have been filtered. */
      currentPath = gtk_tree_model_get_path(GTK_TREE_MODEL(treeStoreFiles),
					    &iter);
      nSet = *(gtk_tree_path_get_indices(currentPath) + 1) + 1;
      gtk_tree_path_free(currentPath);
    }
  else
    nSet = 0;
  gtk_tree_model_get(GTK_TREE_MODEL(treeStoreFiles), &iter,
		     COLUMN_NAME, &filename,
		     COLUMN_NAME_UTF8, &utf8,
		     COLUMN_FILE_KIND, &kind, -1);
  res = browserLoad(filename, utf8, kind, iterSelected, nSet);
  g_free(filename);
  g_free(utf8);
  return res;
}

static gboolean setFiles(VisuData *data, const gchar *filename, const gchar *utf8,
			 int fileKind, int nbKind, VisuData *oldData)
{
  int kind;
  gchar *pt, **tokens, *name;

  g_return_val_if_fail(data && fileKind >= 0 && fileKind <= nbKind, FALSE);

  if (fileKind < nbKind)
    {
      /* The case where we just change one filename, using oldData. */
      if (nbKind > 1 && !oldData)
	return FALSE;

      visuDataAdd_file(data, (gchar*)filename, fileKind, (FileFormat*)0);
      for (kind = 0; kind < nbKind; kind++)
	if (kind != fileKind)
	  visuDataAdd_file(data, visuDataGet_file(oldData, kind, (FileFormat**)0),
			   kind, (FileFormat*)0);
      return TRUE;
    }
  else
    {
      /* The case where the name is used for all files. */
      pt = strrchr(utf8, ']');
      *pt = '\0';
      pt = strrchr(utf8, '[');
      tokens = g_strsplit(pt + 1, ",", nbKind);
      for (kind = 0; kind < nbKind; kind++)
	{
	  name = g_strdup_printf("%s.%s", filename, tokens[kind]);
	  visuDataAdd_file(data, name, kind, (FileFormat*)0);
	}
      g_strfreev(tokens);
      return TRUE;
    }
}

gboolean browserLoad(gchar *filename, gchar *utf8, int fileKind,
		     GtkTreeIter *iter, int nSet)
{
  GError *error;
  VisuData *data, *prevData;
  RenderingMethod *method;
  gboolean changeElement, res, valid;
  VisuData *dataObj;
  RenderingWindow *window;
  GtkTreeIter parentIter, childIter, filterIter;
  int iSet, nSets;
  GtkTreePath *path;
  gchar *comment, *buf, *commentFormat;

  g_return_val_if_fail(iter, FALSE);

  window = RENDERING_WINDOW(visuRenderingWindowGet_current());
  g_return_val_if_fail(window, FALSE);

  dataObj = toolPanelGet_visuData(TOOL_PANEL(panelBrowser));
  if (dataObj)
    data = visuDataNew_withOpenGLView(visuDataGet_openGLView(dataObj));
  else
    data = visuDataNew();
  g_return_val_if_fail(data, FALSE);

  method = getRenderingMethodInUse();
  g_return_val_if_fail(method, FALSE);

  if (!setFiles(data, filename, utf8, fileKind,
		renderingMethodGet_nbFileType(method),
		toolPanelGet_visuData(TOOL_PANEL(panelBrowser))))
    {
      visuGtkRaise_warning(_("Loading a file"),
			   _("Can't load this file through the browser because it"
			     " requires to read several files. You should use the 'Open'"
			     " button on the main panel and then use the browser to vary"
			     " one kind of file at a time."), (GtkWindow*)0);
      g_object_unref(data);
      return FALSE;
    }

  prevData = dataObj;

  error = (GError*)0;
  res = visuBasicLoad_dataFromFile(data, (FileFormat*)0, nSet, &error);
  if (!res)
    {
      visuGtkRaise_warning(_("Loading a file"), error->message, (GtkWindow*)0);
      dataObj = (VisuData*)0;
    }
  else
    dataObj = data;

  if (prevData && dataObj)
    {
      changeElement = visuData_compareElements(prevData, dataObj);
      visuDataSet_changeElementFlag(dataObj, changeElement);
      /* Useless unref since, attaching dataObj to the window on
	 next line will unref the previous VisuData object attached. */
/*       g_object_unref(G_OBJECT(prevData)); */
    }
  renderingWindowSet_visuData(window, dataObj);
  if (dataObj)
    {
      /* If the file has more than one node set, we create the child
	 entries in the tree view. */
      nSets = visuDataGet_nSet(dataObj);
      if (nSets > 1)
	{
	  /* We get the parent iter. */
	  if (!gtk_tree_model_iter_parent(GTK_TREE_MODEL(treeStoreFilesFilter),
					  &parentIter, iter))
	    {
	      gtk_tree_model_filter_convert_iter_to_child_iter
		(GTK_TREE_MODEL_FILTER(treeStoreFilesFilter),
		 &parentIter, iter);
	      /* We clear its children. */
	      valid = gtk_tree_model_iter_children
		(GTK_TREE_MODEL(treeStoreFiles), &childIter, &parentIter);
	      while (valid)
		{
		  gtk_tree_store_remove(treeStoreFiles, &childIter);
		  valid = gtk_tree_model_iter_children
		    (GTK_TREE_MODEL(treeStoreFiles), &childIter, &parentIter);
		}
	      /* We create the new ones. */
	      commentFormat = g_strdup_printf(_("data set %s0%dd"), "%",
					      (int)log10(nSets) + 1);
	      for (iSet = 1; iSet < nSets; iSet++)
		{
		  comment = visuDataGet_fileCommentary(data, iSet);
		  if (!comment || !comment[0])
		    {
		      buf = g_strdup_printf(commentFormat, iSet + 1);
		      comment = buf;
		    }
		  else
		    buf = (gchar*)0;
#if GTK_MINOR_VERSION > 9
		  gtk_tree_store_insert_with_values(treeStoreFiles,
						    &childIter, &parentIter, iSet,
						    COLUMN_BOOLEAN, TRUE,
						    COLUMN_NAME, filename,
						    COLUMN_NAME_UTF8, comment,
						    COLUMN_ACTIVE, TRUE,
						    COLUMN_FILE_KIND, fileKind,
						    -1);
#else							
		  gtk_tree_store_insert(treeStoreFiles, &childIter,
					&parentIter, iSet);
		  gtk_tree_store_set(treeStoreFiles, &childIter,
				     COLUMN_BOOLEAN, TRUE,
				     COLUMN_NAME, filename,
				     COLUMN_NAME_UTF8, comment,
				     COLUMN_ACTIVE, TRUE,
				     COLUMN_FILE_KIND, fileKind,
				     -1);
#endif
		  if (buf)
		    g_free(buf);
		}
	      g_free(commentFormat);
	      /* We expand the row. */
	      gtk_tree_model_filter_convert_child_iter_to_iter
		(GTK_TREE_MODEL_FILTER(treeStoreFilesFilter),
		 &filterIter, &parentIter);
	      path = gtk_tree_model_get_path(GTK_TREE_MODEL(treeStoreFilesFilter),
					     &filterIter);
	      gtk_tree_view_expand_row(GTK_TREE_VIEW(fileTree), path, TRUE);
	      gtk_tree_path_free(path);
	    }
	}
      
      /* We render the new data set. */
      visuData_createAllNodes(dataObj);
      g_idle_add_full(G_PRIORITY_HIGH_IDLE, visuObjectRedraw,
		      (gpointer)0, (GDestroyNotify)0);
    }

  if (error)
    g_error_free(error);

  if (dataObj)
    return TRUE;
  else
    return FALSE;
}

gboolean checkFile(GtkTreeModel *model, GtkTreePath *path _U_,
		   GtkTreeIter *iter, gpointer data)
{
  gboolean filterOk;
  GtkTreeIter iterList;
  GtkTreeStore *tree;

  /* If we check the row, the checkbutton is on only
     if the name is in accordance with the filter. */
  gtk_tree_model_filter_convert_iter_to_child_iter(GTK_TREE_MODEL_FILTER(treeStoreFilesFilter),
						   &iterList, iter);
  tree = GTK_TREE_STORE(gtk_tree_model_filter_get_model(GTK_TREE_MODEL_FILTER(treeStoreFilesFilter)));
  if (GPOINTER_TO_INT(data))
    {
      gtk_tree_model_get(GTK_TREE_MODEL(model), iter, COLUMN_ACTIVE,
			 &filterOk, -1);
      
      if (filterOk)
	gtk_tree_store_set(tree, &iterList, COLUMN_BOOLEAN,
			   TRUE, -1);
      else
	gtk_tree_store_set(tree, &iterList, COLUMN_BOOLEAN,
			   FALSE, -1);
    }
  else
    {
      gtk_tree_store_set(tree, &iterList, COLUMN_BOOLEAN,
			 FALSE, -1);
    }
  return FALSE;
}

void panelBrowserSet_currentDirectory(const gchar *dir)
{
  g_return_if_fail(dir && dir[0]);

  /* Test if the new directory is not already the current one. */
  if (currentBrowseredDirectory && currentBrowseredDirectory[0] &&
      !strcmp(currentBrowseredDirectory[0], dir) && !currentBrowseredDirectory[1])
    return;

  /* Set the new directory current. */
  if (currentBrowseredDirectory)
    g_strfreev(currentBrowseredDirectory);
  currentBrowseredDirectory = g_malloc(sizeof(gchar*) * 2);
  currentBrowseredDirectory[0] = g_strdup(dir);
  currentBrowseredDirectory[1] = (gchar*)0;
  DBG_fprintf(stderr, "Panel Browser: set currentBrowseredDirectory to '%s'.\n",
	      currentBrowseredDirectory[0]);
  if (commonBrowseredDirectory)
    g_free(commonBrowseredDirectory);
  commonBrowseredDirectory = normalize_path(dir);
  DBG_fprintf(stderr, "Panel Browser: set common path to '%s'.\n",
	      commonBrowseredDirectory);

  /* Refresh the list if visible, else let the page-enter
     signal do it. */
  if (toolPanelGet_visible(TOOL_PANEL(panelBrowser)))
    panelBrowserBrowse_directory();
  else
    flagListLoaded = FALSE;
}
void panelBrowserSet_currentDirectories(gchar **dirs)
{
  int i, j;
  gchar *tmp;

  g_return_if_fail(dirs && dirs[0]);

  /* Set the new directory current. */
  if (currentBrowseredDirectory)
    g_strfreev(currentBrowseredDirectory);
  currentBrowseredDirectory = dirs;

  DBG_fprintf(stderr, "Panel Browser: set currentBrowseredDirectory to the list:.\n");
  /* Try to find a common path for all directories. */
  if (commonBrowseredDirectory)
    g_free(commonBrowseredDirectory);
  commonBrowseredDirectory = g_strdup(dirs[0]);
  for (i = 0; dirs[i]; i++)
    {
      DBG_fprintf(stderr, " | '%s'\n", dirs[i]);
      for (j = 0; dirs[i][j] && commonBrowseredDirectory[j]; j++)
	if (dirs[i][j] != commonBrowseredDirectory[j])
	  commonBrowseredDirectory[j] = '\0';
    }
  tmp = commonBrowseredDirectory;
  commonBrowseredDirectory = normalize_path(tmp);
  g_free(tmp);
  DBG_fprintf(stderr, "Panel Browser: set common path to '%s'.\n",
	      commonBrowseredDirectory);

  /* Refresh the list if visible, else let the page-enter
     signal do it. */
  if (toolPanelGet_visible(TOOL_PANEL(panelBrowser)))
    panelBrowserBrowse_directory();
  else
    flagListLoaded = FALSE;
}

static void associateFiles(int nbKind, int commonPathLen)
{
  gboolean valid;
  GtkTreeIter iter;
  int kind, searchKind, i;
  gchar *filename, *searchName, *pt, *fileUTF8;
  GList *tmpLst, *storeLst;
  gsize lu, ecrit;
  gchar **ext;

  /* Read all the stored files and try to associate them on their names.
     When the name matches the previous one without the extension, then the kind
     is stored. If a full set of names can be retrieved, then it is stored. The
     extensions are stored in the ext array. */
  searchName = g_strdup(".");
  searchKind = 0;
  storeLst = (GList*)0;
  ext = g_malloc(sizeof(gchar*) * (nbKind + 1));
  memset(ext, 0, sizeof(gchar*) * (nbKind + 1));
  for (valid = gtk_tree_model_get_iter_first(GTK_TREE_MODEL(treeStoreFiles), &iter);
       valid; valid = gtk_tree_model_iter_next(GTK_TREE_MODEL(treeStoreFiles), &iter))
    {
      gtk_tree_model_get(GTK_TREE_MODEL(treeStoreFiles), &iter,
			 COLUMN_FILE_KIND, &kind, COLUMN_NAME, &filename, -1);
      pt = strrchr(filename, '.');
      if (pt)
	*pt = '\0';
      if (kind >= 0 && kind < nbKind)
	ext[kind] = g_strdup(pt + 1);
/*       fprintf(stderr, "'%s' %d %d\n", filename, kind, searchKind); */
      if (kind >= 0 && kind < nbKind && !strcmp(searchName, filename))
	{
	  /* Ok, match, we continue. */
	  searchKind += kind;
	  g_free(filename);
	  if (searchKind == nbKind * (nbKind - 1) / 2)
	    {
	      pt = g_strjoinv(",", ext);
	      /* We found a complete set. */
	      storeLst =
		g_list_prepend(storeLst, g_markup_printf_escaped("%s<span size=\"smaller\" foreground=\"grey\">.[%s]</span>", searchName, pt));
	      g_free(pt);
	      for (i = 0; i < nbKind; i++)
		g_free(ext[i]);
	      memset(ext, 0, sizeof(gchar*) * (nbKind + 1));
	    }
	}
      else
	{
	  /* Doesn't match, try for new. */
	  g_free(searchName);
	  for (i = 0; i < nbKind; i++)
	    if (ext[i])
	      g_free(ext[i]);
	  memset(ext, 0, sizeof(gchar*) * (nbKind + 1));
	  searchName = filename;
	  searchKind = kind;
	  if (kind >= 0 && kind < nbKind)
	    ext[kind] = g_strdup(pt + 1);
	}
    }
  g_free(searchName);
  /* Store now the associated list. */
  for (tmpLst = storeLst; tmpLst; tmpLst = g_list_next(tmpLst))
    {
      fileUTF8 = g_filename_to_utf8((gchar*)tmpLst->data + commonPathLen + 1,
				    -1, &lu, &ecrit, NULL);
      pt = g_strrstr((gchar*)tmpLst->data, "<span ");
      *pt = '\0';
/*       fprintf(stderr, "Panel Browser: found association for '%s'.\n", fileUTF8); */
#if GTK_MINOR_VERSION > 9
      gtk_tree_store_insert_with_values(treeStoreFiles, &iter, (GtkTreeIter*)0, 0,
					COLUMN_BOOLEAN, FALSE,
					COLUMN_NAME, tmpLst->data,
					COLUMN_NAME_UTF8, fileUTF8,
					COLUMN_ACTIVE, TRUE,
					COLUMN_FILE_KIND, nbKind,
					-1);
#else							
      gtk_tree_store_prepend(treeStoreFiles, &iter, (GtkTreeIter*)0);
      gtk_tree_store_set(treeStoreFiles, &iter,
			 COLUMN_BOOLEAN, FALSE,
			 COLUMN_NAME, tmpLst->data,
			 COLUMN_NAME_UTF8, fileUTF8,
			 COLUMN_ACTIVE, TRUE,
			 COLUMN_FILE_KIND, nbKind,
			 -1);
#endif
      g_free(tmpLst->data);
      if (fileUTF8)
	g_free(fileUTF8);
    }
  g_list_free(storeLst);
}

static void panelBrowserBrowse_directory()
{
  RenderingMethod* method;
  GList *tmpLst, *tmpLst2;
  GDir *gdir;
  gint nbKind, kind, len;
  int *nbPattern, i, nbFiles;
  GPatternSpec ***pattern;
  gchar nbFilesLabel[36];
#if GLIB_MINOR_VERSION > 5
  const gchar **charset;
#else
  const gchar *charset;
#endif

  if (!currentBrowseredDirectory)
    return;

  method = getRenderingMethodInUse();
  if (!method)
    return;

  DBG_fprintf(stderr, "Panel browser : cleaning of the list.\n");
  gtk_tree_store_clear(treeStoreFiles);
  flagListLoaded = FALSE;
  
  nbKind = renderingMethodGet_nbFileType(method);
  pattern = g_malloc(sizeof(GPatternSpec**) * nbKind);
  nbPattern = g_malloc(sizeof(int) * nbKind);
  for( kind = 0; kind < nbKind; kind++)
    {
      nbPattern[kind] = 0;
      tmpLst = renderingMethodGet_fileType(method, kind);
      while (tmpLst)
	{
	  nbPattern[kind] += g_list_length(((FileFormat*)tmpLst->data)->fileType);
	  tmpLst = g_list_next(tmpLst);
	}
      if (nbPattern[kind])
	{
	  pattern[kind] = g_malloc(sizeof(GPatternSpec*) * nbPattern[kind]);
	  i = 0;
	  tmpLst = renderingMethodGet_fileType(method, kind);
	  while (tmpLst)
	    {
	      tmpLst2 = ((FileFormat*)tmpLst->data)->fileType;
	      while (tmpLst2)
		{
		  pattern[kind][i] = g_pattern_spec_new((char*)tmpLst2->data);
		  tmpLst2 = g_list_next(tmpLst2);
		  i++;
		}
	      tmpLst = g_list_next(tmpLst);
	    }
	}
      else
	{
	  g_warning("WARNING! The present rendering method has no"
		    " pattern for prefered file of kind %d, assuming '*'.\n", kind);
	  pattern[kind] = g_malloc(sizeof(GPatternSpec*));
	  nbPattern[kind] = 1;
	  pattern[kind][0] = g_pattern_spec_new((char*)"*");
	}
    }

  flagWorkInProgress = TRUE;
  flagBarIsShown = FALSE;
  gtk_tree_view_set_model(GTK_TREE_VIEW(fileTree),
			  (GtkTreeModel*)0);

  nbFiles = 0;
  sprintf(nbFilesLabel, _("%d files found."), nbFiles);

  /* Add a timeout to show the progress bar if the scan is too long. */
  g_timeout_add(150, showProgressBar, (gpointer)nbFilesLabel);

#if GLIB_MINOR_VERSION > 5
  if (g_get_filename_charsets(&charset))
#else
  if (g_get_charset(&charset))
#endif
    len = g_utf8_strlen(commonBrowseredDirectory, -1);
  else
    len = strlen(commonBrowseredDirectory);
  for (i = 0; currentBrowseredDirectory[i]; i++)
    {
      DBG_fprintf(stderr, "Panel browser : scanning directory '%s'.\n",
		  currentBrowseredDirectory[i]);
      gdir = g_dir_open(currentBrowseredDirectory[i], 0, NULL);
      if (!gdir)
	{
	  visuGtkRaise_warning(_("Browsing a directory"),
			       _("The specified directory is unreadable."),
			       (GtkWindow*)0);
	  flagListLoaded = TRUE;
	  for (kind = 0; kind < nbKind; kind++)
	    {
	      for (i = 0; i < nbPattern[kind]; i++)
		g_pattern_spec_free(pattern[kind][i]);
	      g_free(pattern[kind]);
	    }
	  g_free(pattern);
	  g_free(nbPattern);
	  flagWorkInProgress = FALSE;
	  gtk_widget_set_sensitive(scrolledwindow1, TRUE);
	  gtk_widget_hide(hboxProgress);
	  flagBarIsShown = FALSE;
	  return;
	}
      addParsedDirectory(len, currentBrowseredDirectory[i], gdir,
			 gtk_toggle_button_get_active
			 (GTK_TOGGLE_BUTTON(buttonRecurse)),
			 nbKind, nbPattern, pattern, &nbFiles, nbFilesLabel);
      g_dir_close(gdir);
    }
  if (nbKind > 1)
    associateFiles(nbKind, len);
  flagWorkInProgress = FALSE;

  panelBrowserSet_labelCurrentDir();

  flagListLoaded = TRUE;
  for (kind = 0; kind < nbKind; kind++)
    {
      for (i = 0; i < nbPattern[kind]; i++)
	g_pattern_spec_free(pattern[kind][i]);
      g_free(pattern[kind]);
    }
  g_free(pattern);
  g_free(nbPattern);
  gtk_tree_view_set_model(GTK_TREE_VIEW(fileTree),
			  GTK_TREE_MODEL(treeStoreFilesFilter));

  gtk_widget_set_sensitive(scrolledwindow1, TRUE);
  gtk_widget_hide(hboxProgress);
  flagBarIsShown = FALSE;

  gtk_tree_model_filter_refilter(GTK_TREE_MODEL_FILTER(treeStoreFilesFilter));
}

static void addParsedDirectory(int commonPathLen, const gchar *root,
			       GDir *gdir, gboolean recurse,
			       int nbKind, int *nbPattern, GPatternSpec ***pattern,
			       int *nbFiles, gchar nbFilesLabel[36])
{
  const gchar *fileFromDir;
  gchar *fileUTF8, *file;
  gboolean passed;
  gsize lu, ecrit;
  GDir *recursedDir;
  gint kind;
  int i;
  GtkTreeIter iter;

  g_return_if_fail(root && root[0]);
  DBG_fprintf(stderr, "Panel Browser: read dir '%s' %d.\n", root, (int)recurse);
  fileFromDir = g_dir_read_name(gdir);
  while(fileFromDir && flagWorkInProgress)
    {
      file = g_build_filename(root, fileFromDir, NULL);
      fileUTF8 = g_filename_to_utf8(file + commonPathLen + 1,
				    -1, &lu, &ecrit, NULL);

      if(fileUTF8)
	{
	  if (g_file_test(file, G_FILE_TEST_IS_DIR))
	    {
	      DBG_fprintf(stderr, "Panel Browser: read dir '%s'\n", file);
	      if (recurse)
		{
		  recursedDir = g_dir_open(file, 0, NULL);
		  if (recursedDir)
		    {
		      addParsedDirectory(commonPathLen, file, recursedDir, TRUE,
					 nbKind, nbPattern, pattern,
					 nbFiles, nbFilesLabel);
		      g_dir_close(recursedDir);
		    }
		}
	    }
	  else
	    {
	      passed = FALSE;
	      for (kind = 0; kind < nbKind && !passed; kind++)
		{
		  for (i = 0; i < nbPattern[kind] && !passed; i++)
		    passed = passed || g_pattern_match_string(pattern[kind][i],
							      fileUTF8);
		}
	      kind -= 1;
	      DBG_fprintf(stderr, "Panel Browser: read file '%s' -> %d\n",
			  file, kind);
#if GTK_MINOR_VERSION > 9
	      gtk_tree_store_insert_with_values(treeStoreFiles, &iter,
						(GtkTreeIter*)0, 0,
						COLUMN_BOOLEAN, FALSE,
						COLUMN_NAME, file,
						COLUMN_NAME_UTF8, fileUTF8,
						COLUMN_ACTIVE, TRUE,
						COLUMN_FILE_KIND, (passed)?kind:-1,
						-1);
#else							
	      gtk_tree_store_append(treeStoreFiles, &iter, (GtkTreeIter*)0);
	      gtk_tree_store_set(treeStoreFiles, &iter,
				 COLUMN_BOOLEAN, FALSE,
				 COLUMN_NAME, file,
				 COLUMN_NAME_UTF8, fileUTF8,
				 COLUMN_DATA, "??",
				 COLUMN_ACTIVE, TRUE,
				 COLUMN_FILE_KIND, (passed)?kind:-1,
				 -1);
#endif
	      *nbFiles += 1;
	      if (flagBarIsShown)
		sprintf(nbFilesLabel, _("%d files found."), *nbFiles);
	    }
	  g_free(fileUTF8);
	}
      g_free(file);
      visuGtkWait();
      fileFromDir = g_dir_read_name(gdir);
    }
}

static void onParseAbortClicked(GtkButton *button _U_, gpointer data _U_)
{
  flagWorkInProgress = FALSE;
  gtk_widget_set_sensitive(scrolledwindow1, TRUE);
  gtk_widget_hide(hboxProgress);
  flagBarIsShown = FALSE;
}


static void onBrowserEnter(ToolPanel *toolPanel _U_, gpointer data _U_)
{
  DBG_fprintf(stderr, "Panel Browser: enter, parse directory (%d).\n",
	      !flagBarIsShown);
  if (!flagListLoaded)
    panelBrowserBrowse_directory();
}

static void onDirectoryClicked(GtkButton *button _U_, gpointer data _U_)
{
  gchar **filenames;

  filenames = visuGtkGet_selectedDirectory
    (toolPanelGet_containerWindow(TOOL_PANEL(panelBrowser)),
     TRUE, commonBrowseredDirectory);

  if (!filenames)
    return;

  panelBrowserSet_currentDirectories(filenames);
}

static void onRecurseToggled(GtkToggleButton *toggle _U_, gpointer data _U_)
{
  panelBrowserBrowse_directory();
}

gboolean pulseProgressBar(gpointer data)
{
  gtk_progress_bar_pulse(GTK_PROGRESS_BAR(progressbar));
  gtk_progress_bar_set_text(GTK_PROGRESS_BAR(progressbar), (gchar*)data);
  return flagWorkInProgress;
}
static gboolean showProgressBar(gpointer data)
{
  if (flagWorkInProgress)
    {
      flagBarIsShown = TRUE;
      gtk_widget_set_sensitive(scrolledwindow1, FALSE);
      gtk_widget_show(hboxProgress);
      DBG_fprintf(stderr, "Panel Browser : scanning is too slow, showing progress bar.\n");
      g_timeout_add(100, pulseProgressBar, (gpointer)data);
    }
  return FALSE;
}


/***************************/
/* Miscellaneous functions */
/***************************/

gboolean gtk_tree_model_get_iter_last(GtkTreeModel *model, GtkTreeIter *last, GtkTreePath **path)
{
  gboolean valid;
  gint n;

  g_return_val_if_fail(model && last, FALSE);

  n = gtk_tree_model_iter_n_children(model, (GtkTreeIter*)0);
  if (n == 0)
    return FALSE;

  valid = gtk_tree_model_iter_nth_child(model, last, (GtkTreeIter*)0, n - 1);
  g_return_val_if_fail(valid, FALSE);

  if (path)
    *path = gtk_tree_model_get_path(model, last);
  
  return TRUE;
}

gboolean panelBrowserIsIterVisible(GtkTreeModel *model, GtkTreeIter *iter,
				   gpointer data _U_)
{
  gboolean passUserFilter;
  gint kind;
  
  gtk_tree_model_get(model, iter,
		     COLUMN_FILE_KIND, &kind,
		     COLUMN_ACTIVE, &passUserFilter,
		     -1);
  return (passUserFilter && (currentComboFilterValue < 0 ||
			     kind == currentComboFilterValue));
}

void onComboFilterChanged(GtkComboBox *combo _U_, gpointer data _U_)
{
  GtkTreeIter comboIter;
  gboolean validIter;

  validIter = gtk_combo_box_get_active_iter(GTK_COMBO_BOX(panelBrowserComboFilter),
					    &comboIter);
  if (!validIter)
    return;

  gtk_tree_model_get(GTK_TREE_MODEL(panelBrowserListFilter), &comboIter,
		     PANEL_BROWSER_COLUMN_FILTER_ID, &currentComboFilterValue,
		     -1);
  gtk_tree_model_filter_refilter(GTK_TREE_MODEL_FILTER(treeStoreFilesFilter));
}

void panelBrowserSet_labelCurrentDir()
{
  gchar *directoryUTF8, *markup;

  if (!commonBrowseredDirectory)
    return;

  directoryUTF8 = g_filename_to_utf8(commonBrowseredDirectory, -1, NULL, NULL, NULL);
  g_return_if_fail(directoryUTF8);

  markup = g_markup_printf_escaped(_("<span style=\"italic\" size=\"smaller\">%s</span>"),
				   directoryUTF8);
  g_free(directoryUTF8);
  gtk_label_set_markup(GTK_LABEL(labelDirectory), markup);
  g_free(markup);
}
